#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <stdint.h>

typedef uint32_t IEEE_FLOAT_DOUBLE;

#define SIG_POS 32
#define EXP_POS 31
#define SIGN_POS 23

#define SIG_SHIFT (SIG_POS-1)
#define EXP_SHIFT (EXP_POS-8)

#define SIG_MASK (uint32_t)pow(2,31)
#define EXP_MASK (0xFF << EXP_SHIFT)
#define SIGN_MASK 0x7FFFFF
#define EXT_SIGN_MASK 0xFFFFFF
#define TWELVE_BIT_MASK 0x0FFF
#define HIDDEN_ONE_MASK (uint32_t)pow(2,23)

char * print_bits(uint32_t bits, char buff[24]){

	int i;
	int msb;
	msb = pow(2,32);
	
	for (i=0;i<23;i++){
		buff[i] = ( (bits&msb) ? '1' : '0' );
		bits = bits << 1;
	}
	/* and find last 1 */
	i=22;
	while (buff[i]=='0') i--;
	if (i<0) i=0;	
	buff[i+1]=0;
	return buff;
}

char *print_hex(uint32_t x, char buff[9]){

	int i;
	int digit;

	for(i = 0; i < 8; i++) {
    		digit = (x >> ((7-i) * 4)) & 0xF;
    		if (digit < 10) buff[i] = ('0' + digit);
		else buff[i] = ('A' + (digit-10));
  	}
	buff[8] = 0;
	return buff;

}


IEEE_FLOAT_DOUBLE string_to_fdouble(char *input){

  	uint32_t sig  = 0;
	int exp  = 0;
	int sign = 0;
 	char buff[24], buff2[9];
	char *p_rear = input + strlen(input)-1;
	uint32_t int_total	   = 0;
	uint32_t float_total = 0;
	uint32_t return_f	   = 0;

	// look for a + or - char 
 	if (*input == '-' || *input == '+')
    	{
      		if (*input =='-' ) sig=1;
      		input++;
    	}
   	// scan the input from left to right as long as there are numbers 
    	while (isdigit (*input))
	{
		int_total *= 10;
		int_total += (uint)(*input-'0');
		input++;
	}
	
	if (*input == '.')
    	{
      		/* scan from the end forward */
      		while (isdigit (*p_rear))
		{
			float_total += (uint) ((*p_rear-'0')<<28);
			float_total /= 10;
			p_rear--;
		}
	}
	if (p_rear!=input) printf("There went something wrong while scanning!");
	
	/* now we have the float and int part -> put them together */
	/* sign-bit */
	return_f |= (sig << SIG_SHIFT);
	
	/* exponent */
	while ( (int_total>>exp) > 0 ) exp++;
	exp--;
	
	/* shift real part into right place */
	return_f |= (int_total << (EXP_SHIFT - exp) );
	
	/* shift frac part */
	return_f |= (float_total >> (4+exp+1) );
	 
	/* and remove hidden one from bit 24 */
	return_f &= ~HIDDEN_ONE_MASK;
	
	/* shift exp */
	return_f |= ( (exp+127) << EXP_SHIFT);
	
	/* print */
	printf("sig:%d,",sig);
	printf("exp:%d; ",exp);
	printf("real int-part:%d,",int_total);
	printf("real float-part:0.%s,",(char*)print_bits(float_total<<4,buff));
	printf("Hex:0x%s\n",print_hex(return_f,buff2));
	
	return return_f;	
    
  
}

void fdouble_to_string(IEEE_FLOAT_DOUBLE in){
  char buff2[9];
  int i=0;
  IEEE_FLOAT_DOUBLE t = in;

  uint32_t sig,exp,sign,int_total;
  float float_total=0.0,run_float=1.0,whole;
  
  sig = ( (in & SIG_MASK) >> SIG_SHIFT);
  exp = ( (in & EXP_MASK) >> EXP_SHIFT) - 127;
  in |= HIDDEN_ONE_MASK;
  int_total = ( (in&EXT_SIGN_MASK) >> (EXP_SHIFT-exp) );
  
  //add float part together
//again - assignement doesn't say how we should do this.. so I am using the old sk00l way, and put it all into a "normal" float
  for (i=0;i<(23-exp);i++){
    run_float /= 2;
    if ( (in << (8+1+exp+i)) & SIG_MASK ) float_total += run_float; 
  }
  //add together whole number  
  whole = (int_total + float_total) * (pow(-1,sig));
  
  /* print */
  printf("sig:%d,",sig);
  printf("exp:%d; ",exp);
  printf("real int-part:%d,",int_total);
  printf("real float-part:%f,",float_total);
  printf("Hex:0x%s\n",print_hex(t,buff2));
  printf("Which is: %f\n",whole);
}

IEEE_FLOAT_DOUBLE multiply(IEEE_FLOAT_DOUBLE a, IEEE_FLOAT_DOUBLE b){

  uint32_t A,B,C,D;
  uint32_t E,F,res,exp;
  char buff[9];
  
  printf("\nMultiplying...\n");
  A = ( ((a|HIDDEN_ONE_MASK) >> 12) & TWELVE_BIT_MASK );
  printf("A: %d\n",A);  
  B = ( a & TWELVE_BIT_MASK );
  printf("B: %d\n",B);  
  C = ( ((b|HIDDEN_ONE_MASK) >> 12 ) & TWELVE_BIT_MASK );
  printf("C: %d\n",C);  
  D = ( b & TWELVE_BIT_MASK );
  printf("D: %d\n",D);  
  E = ( (B*D) >> 12 );
  printf("E: %d\n",E);  
  F = ( ((A*D)+(B*C)+E) >> 12);
  printf("F: %d\n",F);  
  
  res = (A*C)+F;
  printf("res: %d\n",res);  
 
  //get the exponent + 1 
  exp = ( (((a&EXP_MASK)>>EXP_SHIFT) - 0x7F) + ((b&EXP_MASK)>>EXP_SHIFT) ) + 1;
  printf("Exponent be4 norm'ing: %d\n",(exp-127));

  //normalize the number (shift 1 to bit 24)
  while (!(res & HIDDEN_ONE_MASK)){
	res <<= 1;
	exp--;
  }
  printf("Exponent after norm'ing: %d\n",(exp-127));

  /* and remove hidden one from bit 24 */
  res &= ~HIDDEN_ONE_MASK;
  printf("res-without-hidden1: %d\n",res);  
  
  //shift the exponent into right place	
  res |= (exp<<EXP_SHIFT);
  
  // get the sign
  res |= ( (a&SIG_MASK) ^ (b&SIG_MASK) );
  
  return res; 
}

int main(){
	char string_a[20],string_b[20];
	char buff[9];
	IEEE_FLOAT_DOUBLE float_a,float_b,float_c;
	
	//it does not say, that we aren't allowed to use "scanf" (in the assignement) - so Im using it here
	printf("Please enter first and second number delimited by <space>\n");
	scanf("%s %s",&string_a,&string_b);
	printf("First: %s,  Second: %s\n",string_a,string_b);
  
	//Convert into floating.point numbers	
        float_a = string_to_fdouble(string_a);
	float_b = string_to_fdouble(string_b);

	//and multiply
	float_c = multiply(float_a,float_b);
	printf("\nResult of Multiplication is:\n");
	fdouble_to_string(float_c);

return 0;
}
