#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <stdint.h>
#include <assert.h>

//global defines
#define REGFILE_SIZE		1024
#define INSTR_MEM_SIZE	1024
#define DATA_MEM_SIZE	1024
#define PROG_START		0
#define __INPUT__
#define __NEAT__

/* some enums */
enum formats 	{UNDEF,R_FORM, I_FORM, J_FORM};
enum r_funcs 	{R_ADD=32,R_SUB=34,R_AND=36,R_OR=37,R_SLT=42};
enum i_ops	{I_BEQ=4,I_ADDI=8,I_ADDIU=9,I_LW=35,I_SW=43};
enum j_ops	{J_J=2};
enum alu_ops	{ALU_AND=0,ALU_OR=1,ALU_ADD=2,ALU_SUB=6,ALU_SLT=7,ALU_NOR=12};

/* typedefs */
typedef uint32_t	Instruction;
typedef uint32_t	Address;
typedef uint8_t	RegNo;		//should be 5 bits
typedef uint32_t	Data;
typedef int16_t	AddImm;
typedef uint32_t	JumpAddr;		//should be 26

typedef uint8_t	ALU_C_Op;		//should be 4bit
typedef uint8_t	OpCode;		//should be 6 bits
typedef uint8_t	Function;		//should be 6 bits
typedef enum {false=0,true=1} bool;
/*
typedef enum formats Formats;
typedef enum r_funcs R_Functions;
typedef enum i_ops I_OPCodes;
typedef enum j_ops J_OPCodes;
*/

/* defines for some const's */

//shifts
#define OP_SHIFT 26
#define RS_SHIFT 21
#define RT_SHIFT 16
#define RD_SHIFT 11
#define SH_SHIFT 6
#define FC_SHIFT 0

//masks
#define OP_MASK (uint32_t)(0x3F<<OP_SHIFT)
#define RS_MASK (uint32_t)(0x1F<<RS_SHIFT)
#define RT_MASK (uint32_t)(0x1F<<RT_SHIFT)
#define RD_MASK (uint32_t)(0x1F<<RD_SHIFT)
#define SH_MASK (uint32_t)(0x1F<<SH_SHIFT)
#define FC_MASK (uint32_t)(0x3F<<FC_SHIFT)
//for I-Format
#define IM_MASK 0x0000FFFF
//for J-Format
#define TA_MASK 0x03FFFFFF
#define TA_SIGN_MASK (uint32_t)(pow(2,25))

//detection bit masks
#define R_FORMAT_DETECT	(uint32_t)(0x10<<OP_SHIFT)		//bin (010 000)
#define I_FORMAT_DETECT	(uint32_t)(0x2c<<OP_SHIFT)		//bin (101 100)
#define J_FORMAT_DETECT	(uint32_t)(0x02<<OP_SHIFT)		//bin (000 010)

//handy macros (stolen from solution to hw1 and 2 ;)
#define GETFIELD(val, mask, shift) (((val) & (mask)) >> (shift))

#define GETOP(val) GETFIELD(val, OP_MASK, OP_SHIFT)
#define GETRS(val) GETFIELD(val, RS_MASK, RS_SHIFT)
#define GETRT(val) GETFIELD(val, RT_MASK, RT_SHIFT)
#define GETRD(val) GETFIELD(val, RD_MASK, RD_SHIFT)
#define GETSH(val) GETFIELD(val, SH_MASK, SH_SHIFT)
#define GETFC(val) GETFIELD(val, FC_MASK, FC_SHIFT)
#define GETIM(val) (AddImm)((val) & (IM_MASK))
#define GETTA(val) ((val) & (TA_MASK))
//  macroes for byte-to-word conversions
#define BYTETOWORD(byteaddr) ((byteaddr) >> 2)
#define WORDTOBYTE(wordaddr) ((wordaddr) << 2)


/* ---------------------------------------------------------------------------- */

/* represents the PC
*  params: 	*addr	-	Address
*		write 	-	Write PC?
* rets:   	none
*/
void PC(Address *addr, bool write)
{
	static Address pc = PROG_START;
	
	printf("PC-Mem:\n<in>  addr=%08x,write=%i\n",*addr,write);
	
	if (write)		pc = *addr;
	else			*addr = pc;

	printf("<out> addr=%08x\n",*addr);
}


/* represents the Registers as a registerfile
*  params: 	rR1 -	first reg to read
*		rR2 -	second reg to read
*		wR -	Reg to write to
*		wD -	Data to write
*		rw -	regWrite switch
*		*rD1 -	first read data from reg
*		*rD2 -	2nd read data from reg
*  rets:   	none
*/
void RegisterFile(RegNo rR1, RegNo rR2, RegNo wR, Data wD, bool rw, Data *rD1, Data *rD2)
{
	static Data registers[REGFILE_SIZE] = {0};

	printf("Reg-File:\n<in>  rR1=%08x,rR2=%08x,wR=%08x,RegW=%i,wD=%08x\n",rR1,rR2,wR,rw,wD);
	
	assert(rR1 < REGFILE_SIZE);
	assert(rR2 < REGFILE_SIZE);
	assert(wR < REGFILE_SIZE);
	
	//write register if requested
	if (rw) registers[wR] = wD;
	//read first reg
	*rD1 = registers[rR1];
	//read 2nd reg
	*rD2 = registers[rR2];
	
	printf("<out> rD1=%08x,rD2=%08x\n",*rD1,*rD2);
}
#ifdef __INPUT__	
/* represents the Instruction Mem as an Array
*  params: 	addr	-	Address to read or store to
*		*data 	-	read instruction
*  rets:   	none
*/
void InstructionMemory(Address addr, Instruction *data, bool write)
{
	static Instruction IMem[INSTR_MEM_SIZE] = {0};

	printf("Instruction-Mem:\n<in>  addr=%08x\n",addr);
	
	assert(BYTETOWORD(addr) < INSTR_MEM_SIZE);

	if (write)	IMem[BYTETOWORD(addr)] = *data;
	else		*data = IMem[BYTETOWORD(addr)];
		
	printf("<out> Instr=%08x\n",*data);
}
#else
/* represents the Instruction Mem as an Array
*  params: 	addr	-	Address to read or store to
*		*data 	-	read instruction
*  rets:   	none
*/
void InstructionMemory(Address addr, Instruction *data)
{
	static Instruction IMem[INSTR_MEM_SIZE] = {0};

	printf("Instruction-Mem:\n<in>  addr=%08x\n",addr);
	
	assert(BYTETOWORD(addr) < INSTR_MEM_SIZE);

	*data = IMem[BYTETOWORD(addr)];
		
	printf("<out> Instr=%08x\n",*data);
}
#endif

/* represents the Data Mem as an Array
*  params: 	addr	-	Address to read or store to
*		mR 	-	Read?
*		mW 	-	Write?
*		*RD 	-	Data read from mem
*		*RW 	-	Data to write to mem
*  rets:   	none
*/
void DataMemory(Address addr, bool mR, bool mW, Data *RD, Data WD)
{
	static Data DMem[DATA_MEM_SIZE] = {0};

	printf("Data-Mem:\n<in>  addr=%08x,MemR=%i,memW=%i,WriteData=%08x\n",addr,mR,mW,WD);
	
	assert(BYTETOWORD(addr) < DATA_MEM_SIZE);
	assert((mR == false) || (mW == false));

	if (mR)		*RD = DMem[BYTETOWORD(addr)];
	else if (mW)	DMem[BYTETOWORD(addr)] = WD;
	
	printf("<out> ReadData=%08x\n",*RD);
}

/* Sign extens the input from 16 to 32 bits
*  params: 	in 	- 	the 16b input
*		*out	-	the 32b output
*  rets:    	none
*/
void SignExtend(AddImm in, Data *out)
{
	printf("Sign-Extend:\n<in>  in=%08x\n",in);
	
	if (in >> 15)	*out = (0xffff << 16) | in;
	else  		*out = (0x0000 << 16) | in;
		
	printf("<out> out=%08x\n",*out);
}


/* Shifts the input left by 2
*  params: 	in 	- 	the 32b input
*		*out	-	the 32b output
*  rets:    	none
*/
void ShiftLeft2(Data in, Data *out)
{
	printf("Shift-Left:\n<in>  in=%08x\n",in);
	
	*out = (in << 2);

	printf("<out> out=%08x\n",*out);
}


/* Mux 2 inputs to 1 depending on the select signal
*  params: 	sel	-	select input (0 = first, 1 = second)
*		in1 	- 	the 1st 32b input
*		in2 	- 	the 2nd 32b input
*		*out	-	the 32b output
*  rets:    	none
*/
void Mux(bool sel, Data in1, Data in2, Data *out)
{
	printf("Mux:\n<in>  SelInput=%i,in1=%08x,in2=%08x\n",sel,in1,in2);
	*out = (sel) ? (in2) : (in1);
	printf("<out> out=%08x\n",*out);
}

/* ANDs just the inputs together (for the branch)
*  params: 	in1 	- 	the 1st input
*		in2 	- 	the 2nd input
*		*out	-	the output
*  rets:    	none
*/
void AND(bool in1, bool in2, bool *out)
{
	printf("AND:\n<in>  in1=%i,in2=%i\n",in1,in2);
	*out = in1 & in2;
	printf("<out> out=%i\n",*out);
}

/* represents the ALU
*  params: 	op	-	the Operation to perform (see enum)
*		in1 	- 	the 1st 32b input
*		in2 	- 	the 2nd 32b input
*		*Zero	-	output zero?
*		*ALUres-	the 32b output
*  rets:    	none
*/
void ALU(ALU_C_Op op, Data in1, Data in2, Data *ALUres, bool *Zero)
{
	printf("ALU:\n<in>  op=%08x,in1=%08x,in2=%08x\n",op,in1,in2);
	switch(op)
	{
		case ALU_AND:
			*ALUres = in1 & in2;
			break;
		case ALU_OR:
			*ALUres = in1 | in2;
			break;
		case ALU_ADD:
			*ALUres = in1 + in2;
			break;
		case ALU_SUB:
			*ALUres = in1 - in2;
			break;
		case ALU_SLT:
			*ALUres = (in1 < in2);
			break;
		case ALU_NOR:
			*ALUres = ~(in1 | in2);
			break;
	}
	*Zero = !(*ALUres);
	printf("<out> res=%08x,zero=%i\n",*ALUres,*Zero);
}

/* represents the Main Control Unit
*  params: 	opc		-	the OpCode from the Instr.
*		*RegDst	- 	rD valid (Register Mux)
*		*ALUSrc 	- 	2nd ALU Input comes from Imm (ALU Mux)
*		*M2R		-	Memory to Register Mux
*		*RegWr	-	Write Signal for Reg
*		*MemR		-	Read Signal for Mem
*		*MemW	-	Write Signal for Mem
*		*Branch	-	Signal ANDed with ALU-Zero for Branch Mux
*		*Jump		-	Signal for Jump
*		*ALUOp1	-	1st AluControl input
*		*ALUOp2	-	2nd AluControl input
*  rets:	none
*/
void Control(OpCode opc, bool *RegDst,  bool *ALUSrc, bool *M2R, bool *RegWr, bool *MemR, bool *MemW, bool *Branch, bool *Jump, bool *ALUOp0, bool *ALUOp1)
{
	printf("Control:\n<in>  opc=%08x\n",opc);
	//R-Format
	if (!opc){
		*RegDst	= true;
		*ALUSrc	= false;
		*M2R		= false;
		*RegWr	= true;
		*MemR	= false;
		*MemW	= false;
		*Branch	= false;
		*Jump	= false;
		*ALUOp1	= true;
		*ALUOp0	= false;
	//LW and SW
	}else if(opc==I_LW || opc==I_SW){
		*RegDst	= false;
		*ALUSrc	= true;
		*M2R		= true;
		*RegWr	=  (opc==I_LW);
		*MemR	=  (opc==I_LW);
		*MemW	= !(opc==I_LW);
		*Branch	= false;
		*Jump	= false;
		*ALUOp1	= false;
		*ALUOp0	= false;
	//ADDI/U
	}else if(opc==I_ADDI || opc==I_ADDIU){
		*RegDst	= false;
		*ALUSrc	= true;
		*M2R		= false;
		*RegWr	= true;
		*MemR	= false;
		*MemW	= false;
		*Branch	= false;
		*Jump	= false;
		*ALUOp1	= false;
		*ALUOp0	= false;
	//BEQ
	}else if (opc==I_BEQ){
		*ALUSrc	= false;
		*RegWr	= false;
		*MemR	= false;
		*MemW	= false;
		*Branch	= true;
		*Jump	= false;
		*ALUOp1	= false;
		*ALUOp0	= true;
	}else if (opc==J_J){
		*ALUSrc	= false;
		*RegWr	= false;
		*MemR	= false;
		*MemW	= false;
		*Branch	= false;
		*Jump	= true;
		*ALUOp1	= false;
		*ALUOp0	= false;
	}
	printf("<out> RegDst=%i, ALUSrc=%i, M2R=%i, RegWr=%i, MemR=%i, MemW=%i, Branch=%i, J=%i, ALUOp1=%i, ALUOp0=%i\n",*RegDst,*ALUSrc,*M2R, *RegWr, *MemR, *MemW, *Branch, *Jump, *ALUOp1, *ALUOp0);
	
}

/* represents the ALU Control Unit
*  params: 	funct		-	the Function from the Instr.
*		ALUOp1	- 	1stALUOp Input
*		ALUOp2 	- 	2nd ALUOp Input
*		*AluContr	-	the ALU Control to feed into the ALU Control Line
* rets:	none
*/
void ALUControl(Function funct, bool ALUOp0, bool ALUOp1, ALU_C_Op *aluc)
{
	printf("ALU-Control:\n<in>  Funct=%08x,ALUOp1=%1d,ALUOp0=%1d\n",funct,ALUOp1,ALUOp0);
	//add
	if (!ALUOp0 & !ALUOp1)
		*aluc = ALU_ADD;
	else if(ALUOp0)
		*aluc = ALU_SUB;
	else if(ALUOp1)
	{
		switch(funct)
		{
			case R_ADD:
				*aluc = ALU_ADD;
				break;
			case R_SUB:
				*aluc = ALU_SUB;
				break;
			case R_AND:
				*aluc = ALU_AND;
				break;
			case R_OR:
				*aluc = ALU_OR;
				break;
			case R_SLT:
				*aluc = ALU_SLT;
				break;
		}
	}
	printf("<out> ALU-Contr=%08x\n",*aluc);
}
#ifdef __INPUT__	
/* Input routine
*  Reads a line at a time from the stdin and puts it in Instruction Memory
*  !!! Is only included if INPUT is defined, coz real Datapath doesnt include this !!!
*
*  params: 	none
*  rets:	none
*/
void read_input()
{
	uint32_t address, command;

	printf("----- reading input into Instr.Mem...(not part of orig. datapath,only convenience) ----\n");
	while (EOF != scanf("%x %x",&address,&command))	
		InstructionMemory(address,&command,true);
	
	printf("---- end of reading input ----\n");
}
#endif

/* Execute routine
*  Is real execution step, loops until end of universe through Instruction Memory
*
*  params: 	none
*  rets:	none
*/
void execute()
{
	/* vars */
	//pc
	Address	pc = PROG_START;
	//Instruction mem
	Instruction 	instr;
	//Control
	bool 		RegDst,ALUSrc,M2R, RegWr, MemR, MemW, Branch, Jump, ALUOp0, ALUOp1;
	//RegisterFile
	RegNo	ReadReg1,ReadReg2,WriteReg;
	Data		ReadRData1,ReadRData2,WriteRData;
	//Sign-Extender
	Data		Immediate;
	//Branch-Shifter
	Data		ImAddr;
	//Jump
	JumpAddr	JAddr;
	JumpAddr	Upper4ofPC;
	//Branch-Adder
	Address	TargetAddr;
	//ALU Mux
	Data		ALUInp2;
	//ALU Control
	ALU_C_Op	aluc;
	//ALU
	bool		ALUZero;
	Data		ALURes;
	//Brancher
	bool		BranchAND;
	Address	NewAddr;
	//Data Mem
	Data		ReadDData, WriteDData;

	
	/* main execution loop (let's get it on) */
	
	while (true){ 
		#ifdef __NEAT__
		printf("----------------Next Instruction---------------------\n");	
		#endif
		//get the PC
		PC(&pc,false);
	
		//call to Instr. mem
		#ifdef __INPUT__
		InstructionMemory(pc,&instr,false);
		#else
		InstructionMemory(pc,&instr);
		#endif
		//and return if no more instructions
		if (!instr) return;
			
		//call to PC-Adder
		ALU(ALU_ADD,pc,4,&pc,&ALUZero);
		
		//call to contr. unit
		Control(GETOP(instr),&RegDst,&ALUSrc,&M2R,&RegWr, &MemR, &MemW, &Branch, &Jump, &ALUOp0, &ALUOp1);
		
		//call to RegFile
		RegisterFile(GETRS(instr), GETRT(instr), GETRD(instr), WriteRData, RegWr, &ReadRData1, &ReadRData2);
		
		//call to Jump-Left-Shifter
		ShiftLeft2(GETTA(instr),&JAddr);
		
		//call to ALU to get the upper 4bits of current PC
		ALU(ALU_AND,pc,0xF0000000,&Upper4ofPC,&ALUZero);
		
		//call to ALU to OR the upper 4 bits of current PC with JumpAddr
		ALU(ALU_OR,JAddr,Upper4ofPC,&JAddr,&ALUZero);
		
		//call to Sign-Extender
		SignExtend(GETIM(instr),&Immediate);
		
		//call to Branch-Left-Shifter
		ShiftLeft2(Immediate,&ImAddr);
		
		//call to Branch-Adder
		ALU(ALU_ADD,ImAddr,pc,&TargetAddr,&ALUZero);
		
		//call to ALU Mux
		Mux(ALUSrc,ReadRData2,Immediate,&ALUInp2);
		
		//call to ALU Control
		ALUControl(GETFC(instr),ALUOp0,ALUOp1,&aluc);
		
		//call to ALU
		ALU(aluc,ReadRData1,ALUInp2,&ALURes,&ALUZero);
		
		//call to Branch-AND and then Branch Mux
		AND(Branch,ALUZero,&BranchAND);
		Mux(BranchAND,pc,TargetAddr,&NewAddr);
		
		//call to Jump MUX
		Mux(Jump,NewAddr,JAddr,&NewAddr);
		
		//call to DataMem
		DataMemory(ALURes,MemR,MemW,&ReadDData,ReadRData2);
		
		//call to Register Mux
		Mux(M2R,ALURes,ReadDData,&WriteRData);
	
		//call to PC Update
		PC(&NewAddr,true);
		
		//call to RegFile
		RegisterFile(GETRS(instr), GETRT(instr), GETRD(instr), WriteRData, RegWr, &ReadRData1, &ReadRData2);
		
	}
	
}
/* main function
*  is called first and then just calls the two functions defined abov, bool *
*  params: 	none
*  rets:	status-code (here:0)
*/
int main(){

#ifdef __INPUT__	
	//read the input into instruction mem
	read_input();
#endif
	
	//and execute
	execute();
		
	return 0;
}
