// Simulation of simple MIPS implementation
// Urrgh, I didn't realize Figure 5.24 was a better starting point for
// you!  Though it's got typoes...
// I'm implementing PC using the technique I'm recommending you use
// for the pipeline registers.
#include <assert.h>
#include <stdint.h>
#include <stdio.h>

// constants
#define INSTMEMSIZE 1024
#define DATAMEMSIZE 1024
#define OPCODEWID 6
#define REGWID 5
#define SHIFTWID 5
#define FUNCWID 6
#define ADDRIMMWID 16
#define JADDRWID 26
#define PCMASK 0xf0000000

#define NUMREGS (1 << REGWID)
#define NUMELEMS(array) (sizeof(array)/sizeof(array[0]))
#define MAXMUX 3

#define SIGN16 0x00008000
#define HIGH16 0xffff0000
#define LOWS16  0x0000ffff

// instruction disassembly
#define MASK(in, size, shift) (((in) >> (shift)) & ((1 << (size)) - 1))
#define FUNC(instr) MASK(instr, FUNCWID, 0)
#define ADDRIMM(instr) MASK(instr, ADDRIMMWID, 0)
#define JADDR(instr) MASK(instr, JADDRWID, 0)
#define RD(instr) MASK(instr, REGWID, FUNCWID+SHIFTWID)
#define RT(instr) MASK(instr, REGWID, FUNCWID+SHIFTWID+REGWID)
#define RS(instr) MASK(instr, REGWID, FUNCWID+SHIFTWID+2*REGWID)
#define OPCODE(instr) MASK(instr, OPCODEWID, FUNCWID+SHIFTWID+3*REGWID)
#define PCHIGHORDER(pc) ((pc) & PCMASK)

// reporting.  We'll leave the old defs in here for debugging
//#define UNIT(name) printf("\n    " #name "\n")
//#define INPUTS() printf("        Inputs:\n");
//#define REPORTIN(name,  digits)     printf("            %s 0x%0" #digits "x\n", #name, name)
//#define OUTPUTS() printf("        Outputs:\n");
//#define REPORTOUT(name, digits)     if ((name) != NULL) printf("            %s 0x%0" #digits "x\n", #name, *name)
#define UNIT(name)
#define INPUTS()
#define REPORTIN(name,  digits)
#define OUTPUTS()
#define REPORTOUT(name, digits)

// helper for memory accesses
#define BYTETOWORD(val) ((val) >> 2)

// types
typedef uint32_t Address;
typedef uint32_t Instruction;
typedef uint32_t Data;
typedef unsigned int OpCode; // really should be six bits
typedef unsigned int RegNo;  // really should be five bits
typedef unsigned int MuxSel; // width really depends on mux type
typedef unsigned int Func;   // really should be six bits

// easier to read control lines when they're enums where appropriate.
typedef enum {FALSE=0, TRUE=1} Bool;
typedef enum {BRNONE=0, BRBEQ=1} BranchType;
typedef enum {JNO=0, JYES=1} JMuxSel;
typedef enum {READ=0, WRITE=1} InstrMemOp;
typedef enum {CTRLADD=0, CTRLSUB=1, CTRLFUNC=2} AluOp;
typedef enum {ALUAND=0, ALUOR=1, ALUADD=2, ALUSUB=6, ALUSLT=7, ALUNOR=0xc} AluCtrl;
typedef enum {FROMRT=0, FROMRD=1} RegDstSel;
typedef enum {FROMREG=0, FROMSE=1, FROMCONST=2} AluSrcSel;
typedef enum {FROMALU=0, FROMMEM=1} MemToRegSel;
typedef enum {FROMADDRINC=0, FROMBRANCHADD=1, FROMJADD=2} BranchMuxSel;

// reporting
void report(uint32_t pc, uint32_t registers[NUMREGS], uint32_t data_memory[DATAMEMSIZE]);

// functional unit forward declarations.

void memOp(Data memory[], unsigned int memsize,
      Address address, Data writeData,
      Bool memRead, Bool memWrite,
      Data *readData);
void alu(uint32_t op1, uint32_t op2, AluCtrl operation, Bool *zero, uint32_t *result);
void control(OpCode opcode, RegDstSel *regDst, JMuxSel *jump, BranchType *branch, Bool *memRead,
             MemToRegSel *memToReg, AluOp *aluOp, Bool *memWrite, AluSrcSel *aluSrc, Bool *regWrite);
void registers(Data regs[NUMREGS], RegNo readReg1, RegNo readReg2, RegNo writeReg, Bool regWrite, Data writeData, Data *readData1, Data *readData2);
void signExtend(Data in, Data *out);
void shiftLeft2(Data in, Data *out);
void jCombine(Address fromPC, Address fromJfield, Address *jumpAddr);
void mux(unsigned int inputs, Data in0, Data in1, Data in2, MuxSel whichone, Data *out);
void aluControl(AluOp aluOp, Func func, AluCtrl *aluControlOut);
void and(Bool in1, Bool in2, Bool *out);
void pc(Address in, Address *out);

//pipeline regs forward decs
void if_id(Address pcIn, Instruction instrIn, Address *pcOut, Instruction *instrOut);
void id_ex(Data idex_readData1, Data idex_readData2,      // from registers to alu (or alu input mux)
    Address idex_pcAluIn,                  // from PC to instruction memory
    Instruction idex_instrMemIn,        // output from instruction memory
    Data idex_signExtendIn,             // from sign extender to alu input mux
    RegDstSel idex_regDstIn,               // is result register selected by rt or rd?
    Bool idex_branchIn,                    // is this a branch instruction?
    JMuxSel idex_jumpIn,                    // is this a branch instruction?
    Bool idex_memReadIn,                   // read data memory?
    MemToRegSel idex_memToRegIn,           // register output from alu or from data memory?
    AluOp idex_aluOpIn,                    // ALU operation
    Bool idex_memWriteIn,                  // write data memory?
    AluSrcSel idex_aluSrcIn,               // second alu input from registers,
    Bool idex_regWriteIn,
//and the outputs *sigh*
    Data *idex_readData1Out, Data *idex_readData2Out,      // from registers to alu (or alu input mux)
    Address *idex_pcAluOut,                  // from PC to instruction memory
    Instruction *idex_instrMemOut,        // output from instruction memory
    Data *idex_signExtendOut,             // from sign extender to alu input mux
    RegDstSel *idex_regDstOut,               // is result register selected by rt or rd?
    Bool *idex_branchOut,                    // is this a branch instruction?
    JMuxSel *idex_jumpOut,                    // is this a branch instruction?
    Bool *idex_memReadOut,                   // read data memory?
    MemToRegSel *idex_memToRegOut,           // register output from alu or from data memory?
    AluOp *idex_aluOpOut,                    // ALU operation
    Bool *idex_memWriteOut,                  // write data memory?
    AluSrcSel *idex_aluSrcOut,               // second alu input from registers,
    Bool *idex_regWriteOut);

void ex_mem(Bool exm_zero,                      // ALU result is 0?
    Data exm_aluResult,                 // result of ALU operation
    Data exm_readData1, Data exm_readData2,      // from registers to alu (or alu input mux)
    Address exm_pcAlu,                  // from PC to instruction memory
    Address exm_branchAddr,          // from branch target adder to branch mux
    Instruction exm_instrMem,        // output from instruction memory
    Address exm_jCombine,            // from jump address combiner to jump mux
    RegDstSel exm_regDst,               // is result register selected by rt or rd?
    Bool exm_branch,                    // is this a branch instruction?
    Bool exm_memRead,                   // read data memory?
    JMuxSel exm_jump,                    // is this a branch instruction?
    MemToRegSel exm_memToReg,           // register output from alu or from data memory?
    Bool exm_memWrite,                  // write data memory?
    Bool exm_regWrite,                  // write result to registers?
//and outputs
	Bool *exm_zeroOut,                      // ALU result is 0?
    Data *exm_aluResultOut,                 // result of ALU operation
    Data *exm_readData1Out, Data *exm_readData2Out,      // from registers to alu (or alu input mux)
    Address *exm_pcAluOut,                  // from PC to instruction memory
    Address *exm_branchAddrOut,          // from branch target adder to branch mux
    Instruction *exm_instrMemOut,        // output from instruction memory
    Address *exm_jCombineOut,            // from jump address combiner to jump mux
    RegDstSel *exm_regDstOut,               // is result register selected by rt or rd?
    Bool *exm_branchOut,                    // is this a branch instruction?
    JMuxSel *exm_jumpOut,                    // is this a branch instruction?
    Bool *exm_memReadOut,                   // read data memory?
    MemToRegSel *exm_memToRegOut,           // register output from alu or from data memory?
    Bool *exm_memWriteOut,                  // write data memory?
    Bool *exm_regWriteOut);
    
void mem_wb(RegDstSel mwb_regDst,               // is result register selected by rt or rd?
    MemToRegSel mwb_memToReg,           // register output from alu or from data memory?
    Bool mwb_regWrite,                  // write result to registers?
    Data mwb_aluResult,                 // result of ALU operation
    Data mwb_readData1, Data mwb_readData2,      // from registers to alu (or alu input mux)
    Data mwb_readData,                  // from data mem to alu/mem mux
    Instruction mwb_instrMem,        // output from instruction memory
    Address mwb_jMux,                // from jump mux to PC
//and outs... :(
    RegDstSel *mwb_regDstOut,               // is result register selected by rt or rd?
    MemToRegSel *mwb_memToRegOut,           // register output from alu or from data memory?
    Bool *mwb_regWriteOut,                  // write result to registers?
    Data *mwb_aluResultOut,                 // result of ALU operation
    Data *mwb_readData1Out, Data *mwb_readData2Out,      // from registers to alu (or alu input mux)
    Data *mwb_readDataOut,                  // from data mem to alu/mem mux
    Instruction *mwb_instrMemOut,        // output from instruction memory
    Address *mwb_jMuxOut);
    

// other forward declarations
void readProgram(Instruction instructionMemory[INSTMEMSIZE]);

int main()
{
    // all the wires (they get more and more...)
/*  -------                                       */	
    //IF
	Address pcOut;                  // from PC to instruction memory
    Address pcAluOut;               // from alu incrementer to jump combiner and branch mux
    Instruction instrMemOut;        // output from instruction memory
    
    // IF/ID
    Address ifid_pcAluOut=0;                  // from PC to instruction memory
    Instruction ifid_instrMemOut=0;        // output from instruction memory
	
	
/*  -------                                       */	
    //ID
    Data signExtendOut;             // from sign extender to alu input mux
    //Control lines
    RegDstSel regDst;               // is result register selected by rt or rd?
    Bool branch;                    // is this a branch instruction?
    JMuxSel jump;                    // is this a branch instruction?
    Bool memRead;                   // read data memory?
    MemToRegSel memToReg;           // register output from alu or from data memory?
    AluOp aluOp;                    // ALU operation
    Bool memWrite;                  // write data memory?
    AluSrcSel aluSrc;               // second alu input from registers,
    Bool regWrite;                  // write result to registers?
    Data readData1, readData2;      // from registers to alu (or alu input mux)
	
   //ID/EX
    Data idex_readData1, idex_readData2;      // from registers to alu (or alu input mux)
    Address idex_pcAluOut;                  // from PC to instruction memory
    Instruction idex_instrMemOut;        // output from instruction memory
    Data idex_signExtendOut;             // from sign extender to alu input mux
    	
	//Control lines
    RegDstSel idex_regDst=0;               // is result register selected by rt or rd?
    JMuxSel idex_jump=0;                    // is this a branch instruction?
    Bool idex_branch;                    // is this a branch instruction?
    Bool idex_memRead;                   // read data memory?
    MemToRegSel idex_memToReg=0;           // register output from alu or from data memory?
    AluOp idex_aluOp=0;                    // ALU operation
    Bool idex_memWrite;                  // write data memory?
    AluSrcSel idex_aluSrc=0;               // second alu input from registers,
    Bool idex_regWrite;                  // write result to registers?
	
/*  -------                                       */	
    //EX
    Address jAddrOut;               // from jump address shifter to jump address combiner
    Address branchShifterOut;       // from branch shifter to branch mux
    Address branchAddrOut;          // from branch target adder to branch mux
    Address jCombineOut;            // from jump address combiner to jump mux
    Data aluMuxOut;                 // from alu input mux to second alu input
    AluCtrl aluControlOut;          // output from alu control unit;
    Bool zero;                      // ALU result is 0?
    Data aluResult;                 // result of ALU operation
    
    //EX/M
    Bool exm_zero;                      // ALU result is 0?
    Data exm_aluResult=0;                 // result of ALU operation
    Data exm_readData1, exm_readData2;      // from registers to alu (or alu input mux)
    Address exm_pcAluOut;                  // from PC to instruction memory
    Address exm_branchAddrOut;          // from branch target adder to branch mux
    Instruction exm_instrMemOut;        // output from instruction memory
    Address exm_jCombineOut;            // from jump address combiner to jump mux
	//Control lines
    RegDstSel exm_regDst=0;               // is result register selected by rt or rd?
    JMuxSel exm_jump=0;                    // is this a branch instruction?
    Bool exm_branch;                    // is this a branch instruction?
    Bool exm_memRead=0;                   // read data memory?
    MemToRegSel exm_memToReg=0;           // register output from alu or from data memory?
    Bool exm_memWrite=0;                  // write data memory?
    Bool exm_regWrite;                  // write result to registers?
    
	
/* -----                       */
    //M
    BranchMuxSel branchMuxControl;  // take a branch?
    Address branchMuxOut;           // from branch mux to jump mux
    Data readData;                  // from data mem to alu/mem mux
    Address jMuxOut;                // from jump mux to PC
				    
    //M/WB
    RegDstSel mwb_regDst=0;               // is result register selected by rt or rd?
    MemToRegSel mwb_memToReg=0;           // register output from alu or from data memory?
    Bool mwb_regWrite=0;                  // write result to registers?
    Data mwb_aluResult;                 // result of ALU operation
    Data mwb_readData1, mwb_readData2;      // from registers to alu (or alu input mux)
    Data mwb_readData;                  // from data mem to alu/mem mux
    Instruction mwb_instrMemOut;        // output from instruction memory
    Address mwb_jMuxOut=0;                // from jump mux to PC
    
	
/*  -------                                       */	
    //WB
    Data aluMemMuxOut;              // from alu/mem mux to register
    Data writeRegMuxOut;            // from write register number mux
    
/*----------------------------------------*/    
    
	// registers and memory
	Data dataMemory[DATAMEMSIZE] = {0};
	Data instructionMemory[INSTMEMSIZE]={0};
	Data regs[NUMREGS] = {0};

    // now the code.

    // read in the program...
    readProgram(instructionMemory);

    // start executing at address 0....
    pcOut = 0x00000000;

    // interestingly, MIPS has no HALT instruction.  So this program
    // should never terminate.  Funny nobody asked me about this....
    // I'm going to put in a hard-loop detector:  if the PC is ever
    // given the same address as the last instruciton executed, I'll
    // end the program.
    while (1) {
		report(pcOut, regs, dataMemory);
		
        // the sequence of these operations must be subject to the
        // constraint that all data is available before it is used.
        // I'm trying to actually call them in the order you'd read
        // them off the figure, going by columns.  This will make
        // pipelining easier later.

        // instruction fetch stage						____IF____
        UNIT(PC Incrementer (alu));
        alu(pcOut, 4, ALUADD, NULL, &pcAluOut);

        UNIT(Instruction Memory);
		memOp(instructionMemory, INSTMEMSIZE, pcOut, 0, TRUE, FALSE, &instrMemOut);
		// detect end-of-program
		if (instrMemOut == 0)
			return 0;
		
		
	/* IF/ID needs		Outputs gonna be
	*	pcALUOut		ifid_pcAluOut,
	*	instrMemOut		ifid_instrMemOut
	*/

        // decode and read registers stage					____ID____
        UNIT(Control Unit);
        control(OPCODE(ifid_instrMemOut),
                &regDst, &jump, &branch, &memRead, &memToReg, &aluOp, &memWrite,
                &aluSrc, &regWrite);
        
        UNIT(Registers);
        registers(regs, RS(ifid_instrMemOut), RT(ifid_instrMemOut), 0, FALSE, 0,
                  &readData1, &readData2);
        
        UNIT(Sign Extender);
        signExtend(ADDRIMM(ifid_instrMemOut), &signExtendOut);
		
	/* ID/EX needs		Outputs
	*	all Control lines: regDst, jump, branch, memRead, memToReg, aluOp, memWrite,aluSrc, regWrite,	Outputs all idex_
	*	pcAluOut		idex_pcAluOut,
	*	instrMemOut		idex_instrMemOut,
	*	signExtendOut	idex_signExtendOut,
	*	readData1		idex_readData1,
	*	readData2		idex_readData2,
	*/

        // alu and branch calc stage						____EX____
        UNIT(Branch left-shifter);
        shiftLeft2(idex_signExtendOut, &branchShifterOut);

        UNIT(Jump left-shifter);
        shiftLeft2(JADDR(idex_instrMemOut), &jAddrOut);

        UNIT(Jump address combiner);
        jCombine(jAddrOut, PCHIGHORDER(idex_pcAluOut), &jCombineOut);

        UNIT(ALU Source Mux);
        mux(3, idex_readData2, idex_signExtendOut, ADDRIMM(idex_instrMemOut), idex_aluSrc, &aluMuxOut);

        UNIT(ALU Control);
        aluControl(idex_aluOp, FUNC(idex_instrMemOut), &aluControlOut);

        UNIT(Branch Target Adder);
        alu(idex_pcAluOut, branchShifterOut, ALUADD, NULL, &branchAddrOut);

        UNIT(Main ALU);
        alu(idex_readData1, aluMuxOut, aluControlOut, &zero, &aluResult);
	
	/* EX/MEM needs		Outputs
	*	Control lines: branch, memRead,memWrite,jump,mem2Reg,RegDst,regWrite
	*	pcAluOut		exm_pcAluOut,
	*	instrMemOut		exm_instrMemOut,
	*	zero			exm_zero,
	*	branchAddrOut	exm_branchAddrOut,
	*	jCombineOut		exm_jCombineOut,
	*	aluResult		exm_aluResult,
	*	readData1		exm_readData1,
	*	readData2		exm_readData2,
	*/

        // branch decision and mem op stage				____MEM____
        UNIT(BEQ And);
        and(exm_branch, exm_zero, &branchMuxControl);

        UNIT(Branch Mux);
        mux(2, exm_pcAluOut, exm_branchAddrOut, 0, branchMuxControl, &branchMuxOut);

        UNIT(Jump mux);
        mux(2, branchMuxOut, exm_jCombineOut, 0, exm_jump, &jMuxOut);

        UNIT(Data Memory);
		memOp(dataMemory, DATAMEMSIZE, exm_aluResult, exm_readData2, exm_memRead, exm_memWrite, &readData);
	
	/* MEM/WB needs		Outputs
	*	Control lines: mem2Reg,RegDst,regWrite
	*	aluResult		mwb_aluResult,
	*	readData		mwb_readData,
	*	instrMemOut		mwb_instrMemOut,
	*	jMuxOut		mwb_jMuxOut,
	*	readData1		mwb_readData1,
	*	readData2		mwb_readData2,
	*/

        // writeback stage							____WB____
        UNIT(ALU/Memory Mux);
        mux(2, mwb_aluResult, mwb_readData, 0, mwb_memToReg, &aluMemMuxOut);

        UNIT(Write Register Mux);
        mux(2, RT(mwb_instrMemOut), RD(mwb_instrMemOut), 0, mwb_regDst, &writeRegMuxOut);

        UNIT(Registers (writeback));
        registers(regs, 0, 0, writeRegMuxOut, mwb_regWrite, aluMemMuxOut, &mwb_readData1, &mwb_readData2);
	
	//make sure we update the PC with the right value even if we are in the first 4 cycles
	if ( (mwb_jMuxOut==0) || ((mwb_jMuxOut+8)==pcAluOut)) mwb_jMuxOut=pcAluOut;
        UNIT(Program Counter);
		pc(mwb_jMuxOut, &pcOut);
		
		
		
	//pipeline regs stage..						____PIPELINE____
	//if_id
	if_id(pcAluOut,instrMemOut,&ifid_pcAluOut,&ifid_instrMemOut);
	
	//id_ex
	id_ex(readData1, readData2,      // from registers to alu (or alu input mux)
		pcAluOut,                  // from PC to instruction memory
		ifid_instrMemOut,        // output from instruction memory
		signExtendOut,             // from sign extender to alu input mux
		regDst,               // is result register selected by rt or rd?
		branch,                    // is this a branch instruction?
		jump,                    // is this a branch instruction?
		memRead,                   // read data memory?
		memToReg,           // register output from alu or from data memory?		
		aluOp,                    // ALU operation
		memWrite,                  // write data memory?
		aluSrc,               // second alu input from registers,
		regWrite,
		//and the outputs *sigh*
		&idex_readData1, &idex_readData2,      // from registers to alu (or alu input mux)
		&idex_pcAluOut,                  // from PC to instruction memory
		&idex_instrMemOut,        // output from instruction memory
		&idex_signExtendOut,             // from sign extender to alu input mux
		&idex_regDst,               // is result register selected by rt or rd?
		&idex_branch,                    // is this a branch instruction?
		&idex_jump,                    // is this a branch instruction?
		&idex_memRead,                   // read data memory?
		&idex_memToReg,           // register output from alu or from data memory?
		&idex_aluOp,                    // ALU operation
		&idex_memWrite,                  // write data memory?
		&idex_aluSrc,               // second alu input from registers,
		&idex_regWrite
	);
	//ex_mem
	ex_mem(zero,                      // ALU result is 0?
		aluResult,                 // result of ALU operation
		idex_readData1, idex_readData2,      // from registers to alu (or alu input mux)
		idex_pcAluOut,                  // from PC to instruction memory
		branchAddrOut,          // from branch target adder to branch mux
		idex_instrMemOut,        // output from instruction memory
		jCombineOut,            // from jump address combiner to jump mux
		idex_regDst,               // is result register selected by rt or rd?
		idex_branch,                    // is this a branch instruction?
		idex_jump,                    // is this a branch instruction?
		idex_memRead,                   // read data memory?
		idex_memToReg,           // register output from alu or from data memory?
		idex_memWrite,                  // write data memory?
		idex_regWrite,                  // write result to registers?
		//and outputs
		&exm_zero,                      // ALU result is 0?
		&exm_aluResult,                 // result of ALU operation
		&exm_readData1, &exm_readData2,      // from registers to alu (or alu input mux)
		&exm_pcAluOut,                  // from PC to instruction memory
		&exm_branchAddrOut,          // from branch target adder to branch mux
		&exm_instrMemOut,        // output from instruction memory
		&exm_jCombineOut,            // from jump address combiner to jump mux
		&exm_regDst,               // is result register selected by rt or rd?
		&exm_branch,                    // is this a branch instruction?
		&exm_jump,                    // is this a branch instruction?
		&exm_memRead,                   // read data memory?
		&exm_memToReg,           // register output from alu or from data memory?
		&exm_memWrite,                  // write data memory?
		&exm_regWrite
	);
	//Mem/WB
	mem_wb(exm_regDst,               // is result register selected by rt or rd?
		exm_memToReg,           // register output from alu or from data memory?
		exm_regWrite,                  // write result to registers?
		exm_aluResult,                 // result of ALU operation
		exm_readData1, exm_readData2,      // from registers to alu (or alu input mux)
		readData,                  // from data mem to alu/mem mux
		exm_instrMemOut,        // output from instruction memory
		jMuxOut,                // from jump mux to PC
		//and outs... :(
		&mwb_regDst,               // is result register selected by rt or rd?
		&mwb_memToReg,           // register output from alu or from data memory?
		&mwb_regWrite,                  // write result to registers?
		&mwb_aluResult,                 // result of ALU operation
		&mwb_readData1, &mwb_readData2,      // from registers to alu (or alu input mux)
		&mwb_readData,                  // from data mem to alu/mem mux
		&mwb_instrMemOut,        // output from instruction memory
		&mwb_jMuxOut
	);

	//End Pipeline...
    }
}

/*****************************************************************************/
/*                                                                           */
/*  readProgram                                                              */
/*                                                                           */
/*      reads program from standard input.                                   */
/*                                                                           */
/*      no parameters                                                        */
/*                                                                           */
/*****************************************************************************/
void readProgram(Instruction instructionMemory[INSTMEMSIZE])
{
    Address PC;
    Instruction InstructionRegister;

    // read in all the instructions....
    while (scanf("%x %x", &PC, &InstructionRegister) != EOF)
        memOp(instructionMemory, INSTMEMSIZE, PC, InstructionRegister, FALSE, TRUE, NULL);

}

/*****************************************************************************/
/*                                                                           */
/*  alu                                                                      */
/*                                                                           */
/*      simulates an ALU.                                                    */
/*                                                                           */
/*      inputs:                                                              */
/*          op1, op2:  alu inputs                                            */
/*          operation: operation to be performed on the inputs               */
/*                                                                           */
/*      outputs:                                                             */
/*          zero:  boolean indicating whether result is 0                    */
/*          result: result of alu operation                                  */
/*                                                                           */
/*      I'm using the same code for all three places that call for an ALU.   */
/*      In two of them, it is just an adder, and I don't have to worry       */
/*      the zero output, but it keeps me from having to write extra code.    */
/*      Because of that, I'm allowing a null pointer on the zero output;     */
/*      rather than aborting on a null pointer there I just don't write to   */
/*      it                                                                   */
/*                                                                           */
/*****************************************************************************/
void alu(uint32_t op1, uint32_t op2, AluCtrl operation, Bool *zero, uint32_t *result)
{
    INPUTS();
    REPORTIN(op1, 8);
    REPORTIN(op2, 8);
    REPORTIN(operation, 1);
    
    assert(result != NULL);

    switch (operation) {
      case ALUAND:
        *result = op1 & op2;
        break;
        
      case ALUOR:
        *result = op1 | op2;
        break;
        
      case ALUADD:
        *result = op1 + op2;
        break;
        
      case ALUSUB:
        *result = op1 - op2;
        break;
        
      case ALUSLT:
        *result = op1 < op2;
        break;
        
      case ALUNOR:
        *result = ~(op1 | op2);
        break;

      default:
        assert(FALSE);
    }

    if (zero != NULL)
        *zero = (*result == 0);

    OUTPUTS();
    REPORTOUT(zero, 1);
    REPORTOUT(result, 8);
}

/*****************************************************************************/
/*                                                                           */
/*  memOp                                                                    */
/*                                                                           */
/*      common code shared between instructionMemory and dataMemory          */
/*      (in C++, this would be a base class...)                              */
/*                                                                           */
/*      inputs:                                                              */
/*          memory:  memory array being  simulated                           */
/*          memsize: size of memory array                                    */
/*          address: address to be read or written                           */
/*          writeData: data to be written                                    */
/*          memRead:  whether or not to read                                 */
/*          memWrite:  whether or not to write                               */
/*                                                                           */
/*      outputs:                                                             */
/*          readData:  data read from memory array                           */
/*                                                                           */
/*      Note that I do check for things like making sure you don't try tod   */
/*      read and write on the same cycle; also, I only check the readData    */
/*      pointer if you are actually going to read.                           */
/*                                                                           */
/*****************************************************************************/
void memOp(Data memory[], unsigned int memsize, Address address,
           Data writeData, Bool memRead, Bool memWrite, Data *readData)
{

    INPUTS();
    REPORTIN(address, 8);
    REPORTIN(writeData, 8);
    REPORTIN(memRead, 1);
    REPORTIN(memWrite, 1);
    
    assert((!memRead && !memWrite) || (BYTETOWORD(address) < memsize));
    assert(!memRead || (readData != NULL));
    assert(!memRead || !memWrite);

    if (memRead)
        *readData = memory[BYTETOWORD(address)];

    if (memWrite)
        memory[BYTETOWORD(address)] = writeData;
    
    OUTPUTS();
    REPORTOUT(readData, 8);
}

/*****************************************************************************/
/*                                                                           */
/*  control                                                                  */
/*                                                                           */
/*      main control unit                                                    */
/*                                                                           */
/*      Input:                                                               */
/*          opcode:  instruction op code                                     */
/*                                                                           */
/*      Outputs:                                                             */
/*          lots of control lines, given the same names as in the text       */
/*                                                                           */
/*      This control unit is actually simpler than the one from hw2; just    */
/*      a great big lookup table.  This assignment didn't actually call      */
/*      for a table that would let us implement the entire instruction set,  */
/*      but it seemed like the simplest way to do it.                        */
/*                                                                           */
/*****************************************************************************/
void control(OpCode opcode, RegDstSel *regDst, JMuxSel *jump, BranchType *branch, Bool *memRead,
        MemToRegSel *memToReg, AluOp *aluOp, Bool *memWrite, AluSrcSel *aluSrc, Bool *regWrite)
{
    const static struct {
        RegDstSel regDst;
        JMuxSel jump;
        BranchType branch;
        Bool memRead;
        MemToRegSel memToReg;
        AluOp aluOp;
        Bool memWrite;
        AluSrcSel aluSrc;
        Bool regWrite;
    } decode[] = {
    // regDst  jump  branch  memRead memToReg aluOp     memWrite aluSrc   regWrite
      {FROMRD, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, TRUE}, //  0 R-Format
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, //  1 
      {FROMRT, JYES, BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, //  2 j
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, //  3
      {FROMRT, JNO,  BRBEQ,  FALSE,  FROMALU, CTRLSUB,  FALSE,   FROMREG, FALSE}, //  4 beq
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, //  5
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, //  6
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, //  7
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLADD,  FALSE,   FROMSE, TRUE}, //  8 addi
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLADD,  FALSE,   FROMCONST, TRUE}, //  9 addiu
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 10
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 11
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 12
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 13
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 14
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 15
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 16
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 17
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 18
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 19
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 20
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 21
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 22
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 23
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 24
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 25
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 26
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 27
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 28
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 29
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 30
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 31
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 32
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 33
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 34
      {FROMRT, JNO,  BRNONE, TRUE,   FROMMEM, CTRLADD,  FALSE,   FROMSE,  TRUE},  // 35 lw
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 36
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 37
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 38
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 39
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 40
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 41
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 42
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLADD,  TRUE,    FROMSE,  FALSE}, // 43 sw
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 44
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 45
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 46
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 47
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 48
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 49
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 50
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 51
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 52
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 53
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 54
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 55
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 56
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 57
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 58
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 59
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 60
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 61
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 62
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}  // 63
    };

    INPUTS();
    REPORTIN(opcode, 2);
    
    assert(opcode < NUMELEMS(decode));
    
    assert(regDst != NULL);
    assert(jump != NULL);
    assert(branch != NULL);
    assert(memRead != NULL);
    assert(memToReg != NULL);
    assert(aluOp != NULL);
    assert(memWrite != NULL);
    assert(aluSrc != NULL);
    assert(regWrite != NULL);

    *regDst   = decode[opcode].regDst;
    *jump     = decode[opcode].jump;
    *branch   = decode[opcode].branch;
    *memRead  = decode[opcode].memRead;
    *memToReg = decode[opcode].memToReg;
    *aluOp    = decode[opcode].aluOp;
    *memWrite = decode[opcode].memWrite;
    *aluSrc   = decode[opcode].aluSrc;
    *regWrite = decode[opcode].regWrite;

    OUTPUTS();
    REPORTOUT(regDst, 1);
    REPORTOUT(jump, 1);
    REPORTOUT(branch, 1);
    REPORTOUT(memRead, 1);
    REPORTOUT(memToReg, 1);
    REPORTOUT(aluOp, 1);
    REPORTOUT(memWrite, 1);
    REPORTOUT(aluSrc, 1);
    REPORTOUT(regWrite, 1);
    
}

/*****************************************************************************/
/*                                                                           */
/*  registers                                                                */
/*                                                                           */
/*      register set                                                         */
/*                                                                           */
/*      inputs:                                                              */
/*          regs:      register file                                         */
/*          readReg1, readReg2:  registers to read                           */
/*          writeReg:  register to write (if any)                            */
/*          regWrite:  whether to try to write a register                    */
/*          writeData: data to write to register (if any)                    */
/*                                                                           */
/*      outputs:                                                             */
/*          readData1, readData2:  data read from registers                  */
/*                                                                           */
/*     This emulates the MIPS register set semantics:  writes happen         */
/*     before reads, and we don't write to $0                                */
/*                                                                           */
/*****************************************************************************/
void registers(Data regs[NUMREGS], RegNo readReg1, RegNo readReg2, RegNo writeReg, Bool regWrite, Data writeData, Data *readData1, Data *readData2)
{

    INPUTS();
    REPORTIN(readReg1, 2);
    REPORTIN(readReg2, 2);
    REPORTIN(writeReg, 2);
    REPORTIN(regWrite, 1);
    REPORTIN(writeData, 8);

    assert(readReg1 < NUMREGS);
    assert(readReg2 < NUMREGS);
    assert(writeReg < NUMREGS);
    assert(readData1 != NULL);
    assert(readData2 != NULL);
    
    if (regWrite && (writeReg != 0))
        regs[writeReg] = writeData;

    *readData1 = regs[readReg1];
    *readData2 = regs[readReg2];

    OUTPUTS();
    REPORTOUT(readData1, 8);
    REPORTOUT(readData2, 8);
}

/*****************************************************************************/
/*                                                                           */
/*  signExtend                                                               */
/*                                                                           */
/*      sign extender                                                        */
/*                                                                           */
/*      input:                                                               */
/*          in:  16 bit value to sign-extend                                 */
/*                                                                           */
/*      output:                                                              */
/*          out: 32 bit sign-extended result                                 */
/*                                                                           */
/*****************************************************************************/
void signExtend(Data in, Data *out)
{
    INPUTS();
    REPORTIN(in, 4);
    assert(out != NULL);

    if (in & SIGN16)
        *out = in | HIGH16;
    else
        *out = in & LOWS16;

    OUTPUTS();
    REPORTOUT(out, 8);
}


/*****************************************************************************/
/*                                                                           */
/*  shiftLeft2                                                               */
/*                                                                           */
/*      left shifter                                                         */
/*                                                                           */
/*      input:                                                               */
/*          in:  32 bit value to be left-shifted by two bits                 */
/*                                                                           */
/*      output:                                                              */
/*          out: 32 bit left-shifted value                                   */
/*                                                                           */
/*****************************************************************************/
void shiftLeft2(Data in, Data *out)
{
    INPUTS();
    REPORTIN(in, 8);
    
    assert(out != NULL);

    *out = in << 2;

    OUTPUTS();
    REPORTOUT(out, 8);
}


/*****************************************************************************/
/*                                                                           */
/*  jCombine                                                                 */
/*                                                                           */
/*      jump address combiner                                                */
/*                                                                           */
/*      input:                                                               */
/*          pc:  high-order four bits of PC (+4)                             */
/*          jaddr: low order 28 bits of jump adder from instruction          */
/*                                                                           */
/*      output:                                                              */
/*          out: jump target address                                         */
/*                                                                           */
/*****************************************************************************/
void jCombine(Address pc, Address jaddr, Address *jumpAddr)
{
    INPUTS();
    REPORTIN(pc, 8);
    REPORTIN(jaddr, 8);

    assert(jumpAddr != NULL);

    *jumpAddr = pc | jaddr;

    OUTPUTS();
    REPORTOUT(jumpAddr, 8);
}

/*****************************************************************************/
/*                                                                           */
/*  mux                                                                      */
/*                                                                           */
/*      two- or three-input mux                                              */
/*                                                                           */
/*      input:                                                               */
/*          inputs:  number of inputs                                        */
/*          in0, in1, in2:  data inputs                                      */
/*          whicone: selector input                                          */
/*                                                                           */
/*      output:                                                              */
/*          out: value passed through mux                                    */
/*                                                                           */
/*                                                                           */
/*       in order to avoid duplicating  code writing several muxes, this     */
/*       is a single mux with a variable number of inputs.  There is a       */
/*       limit on the max number of inputs (set by MAXMUX), which "inputs"   */
/*       is checked against.  whichone is then checked against inputs.       */
/*                                                                           */
/*       The "do a shared mux with custom muxes calling it" (like with       */
/*       the memories) was considered, but would have had too much code in   */
/*       the derived muxes.                                                  */
/*                                                                           */
/*****************************************************************************/
void mux(unsigned int inputs, Data in0, Data in1, Data in2, MuxSel whichone, Data *out)
{
    INPUTS();
    REPORTIN(inputs, 1);
    REPORTIN(in0, 8);
    REPORTIN(in1, 8);
    REPORTIN(in2, 8);
    REPORTIN(whichone, 1);
    
    assert(out != NULL);
    assert(inputs <= MAXMUX);
    assert(whichone < inputs);

    switch (whichone) {
      case 0:
        *out = in0;
        break;

      case 1:
        *out = in1;
        break;

      case 2:
        *out = in2;
        break;
    }

    OUTPUTS();
    REPORTOUT(out, 8);
}
/*****************************************************************************/
/*                                                                           */
/*  aluControl                                                               */
/*                                                                           */
/*     alu control unit                                                      */
/*                                                                           */
/*     inputs:                                                               */
/*         aluOp:  aluOp from main control unit                              */
/*         func:   func field from instruction                               */
/*                                                                           */
/*     output:                                                               */
/*         aluControlOut                                                     */
/*                                                                           */
/*****************************************************************************/
void aluControl(AluOp aluOp, Func func, AluCtrl *aluControlOut)
{
    AluCtrl resFromAluOp[] = {
      ALUADD,  // 00
      ALUSUB,  // 01
      CTRLFUNC // 10
    };
    
    AluCtrl resFromFunc[] = {
      ALUAND, //  0
      ALUAND, //  1
      ALUAND, //  2
      ALUAND, //  3
      ALUAND, //  4
      ALUAND, //  5
      ALUAND, //  6
      ALUAND, //  7
      ALUAND, //  8
      ALUAND, //  9
      ALUAND, // 10
      ALUAND, // 11
      ALUAND, // 12
      ALUAND, // 13
      ALUAND, // 14
      ALUAND, // 15
      ALUAND, // 16
      ALUAND, // 17
      ALUAND, // 18
      ALUAND, // 19
      ALUAND, // 20
      ALUAND, // 21
      ALUAND, // 22
      ALUAND, // 23
      ALUAND, // 24
      ALUAND, // 25
      ALUAND, // 26
      ALUAND, // 27
      ALUAND, // 28
      ALUAND, // 29
      ALUAND, // 30
      ALUAND, // 31
      ALUADD, // 32 add
      ALUAND, // 33
      ALUSUB, // 34 sub
      ALUAND, // 35
      ALUAND, // 36
      ALUAND, // 37
      ALUAND, // 38
      ALUAND, // 39
      ALUAND, // 40
      ALUAND, // 41
      ALUAND, // 42
      ALUAND, // 43
      ALUAND, // 44
      ALUAND, // 45
      ALUAND, // 46
      ALUAND, // 47
      ALUAND, // 48
      ALUAND, // 49
      ALUAND, // 50
      ALUAND, // 51
      ALUAND, // 52
      ALUAND, // 53
      ALUAND, // 54
      ALUAND, // 55
      ALUAND, // 56
      ALUAND, // 57
      ALUAND, // 58
      ALUAND, // 59
      ALUAND, // 60
      ALUAND, // 61
      ALUAND, // 62
      ALUAND  // 63
    };

    INPUTS();
    REPORTIN(aluOp, 1);
    REPORTIN(func, 2);

    assert(aluOp < NUMELEMS(resFromAluOp));
    assert(func < NUMELEMS(resFromFunc));
    assert(aluControlOut != NULL);

    if (aluOp != CTRLFUNC)
        *aluControlOut = resFromAluOp[aluOp];
    else
        *aluControlOut = resFromFunc[func];

    OUTPUTS();
    REPORTOUT(aluControlOut, 2);
}

/*****************************************************************************/
/*                                                                           */
/*  and                                                                      */
/*                                                                           */
/*      and gate                                                             */
/*                                                                           */
/*      inputs:                                                              */
/*          in1, in2                                                         */
/*                                                                           */
/*      outputs:                                                             */
/*          out                                                              */
/*                                                                           */
/*****************************************************************************/
void and(Bool in1, Bool in2, Bool *out)
{
    INPUTS();
    REPORTIN(in1, 1);
    REPORTIN(in2, 1);
    
    assert (out != NULL);

    *out = in1 & in2;

    OUTPUTS();
    REPORTOUT(out, 1);
}

/*****************************************************************************/
/*                                                                           */
/*  pc                                                                       */
/*                                                                           */
/*      program counter register                                             */
/*                                                                           */
/*      input:                                                               */
/*          in                                                               */
/*                                                                           */
/*      outputs:                                                             */
/*          out                                                              */
/*                                                                           */
/*      returns:                                                             */
/*          TRUE if processing should continue, FALSE if the new PC has the  */
/*          same value as the old PC -- gives a way to halt the simulation   */
/*          without adding a bogus "halt" instruction                        */
/*                                                                           */
/*****************************************************************************/
void pc(Address in, Address *out)
{
    INPUTS();
    REPORTIN(in, 8);
    
    assert(out != NULL);
	
    *out = in;

    OUTPUTS();
    REPORTOUT(out, 8);

}
/*******************************PIPELINE REGS*******************************/
/*                                                                           */
/*  IF/ID                                                                    */
/*...................................*/
void if_id(Address pcIn, Instruction instrIn, Address *pcOut, Instruction *instrOut)
{
    *pcOut = pcIn;
    *instrOut = instrIn;
}
/*  ID/EX                                                                    */
/*...................................*/
void id_ex(Data idex_readData1In, Data idex_readData2In,      
    Address idex_pcAluIn,                 
    Instruction idex_instrMemIn,        
    Data idex_signExtendIn,             
    RegDstSel idex_regDstIn,            
    Bool idex_branchIn,                    
    JMuxSel idex_jumpIn,                    
    Bool idex_memReadIn,                 
    MemToRegSel idex_memToRegIn,   
    AluOp idex_aluOpIn,                   
    Bool idex_memWriteIn,                
    AluSrcSel idex_aluSrcIn,              
    Bool idex_regWriteIn,
//and the outputs *sigh*
    Data *idex_readData1Out, Data *idex_readData2Out,      
    Address *idex_pcAluOut,                 
    Instruction *idex_instrMemOut,        
    Data *idex_signExtendOut,            
    RegDstSel *idex_regDstOut,           
    Bool *idex_branchOut,                   
    JMuxSel *idex_jumpOut,                   
    Bool *idex_memReadOut,                
    MemToRegSel *idex_memToRegOut,    
    AluOp *idex_aluOpOut,                   
    Bool *idex_memWriteOut,                
    AluSrcSel *idex_aluSrcOut,              
    Bool *idex_regWriteOut)
{
    *idex_readData1Out 	= idex_readData1In;
    *idex_readData2Out 	= idex_readData2In;      
    *idex_pcAluOut 		= idex_pcAluIn;          
    *idex_instrMemOut 		= idex_instrMemIn;      
    *idex_signExtendOut 	= idex_signExtendIn;    
    *idex_regDstOut		= idex_regDstIn;               
    *idex_branchOut		= idex_branchIn;                
    *idex_jumpOut		= idex_jumpIn;                
    *idex_memReadOut		= idex_memReadIn;
    *idex_memToRegOut	= idex_memToRegIn;           
    *idex_aluOpOut		= idex_aluOpIn;                  
    *idex_memWriteOut		= idex_memWriteIn;             
    *idex_aluSrcOut		= idex_aluSrcIn;               
    *idex_regWriteOut		= idex_regWriteIn;
}
/*  EX/MEM                                                                    */
/*...................................*/
void ex_mem(Bool exm_zero,                      // ALU result is 0?
    Data exm_aluResult,                 // result of ALU operation
    Data exm_readData1, Data exm_readData2,      // from registers to alu (or alu input mux)
    Address exm_pcAlu,                  // from PC to instruction memory
    Address exm_branchAddr,          // from branch target adder to branch mux
    Instruction exm_instrMem,        // output from instruction memory
    Address exm_jCombine,            // from jump address combiner to jump mux
    RegDstSel exm_regDst,               // is result register selected by rt or rd?
    Bool exm_branch,                    // is this a branch instruction?
    JMuxSel exm_jump,                    // is this a branch instruction?
    Bool exm_memRead,                   // read data memory?
    MemToRegSel exm_memToReg,           // register output from alu or from data memory?
    Bool exm_memWrite,                  // write data memory?
    Bool exm_regWrite,                  // write result to registers?
//and outputs
	Bool *exm_zeroOut,                      // ALU result is 0?
    Data *exm_aluResultOut,                 // result of ALU operation
    Data *exm_readData1Out, Data *exm_readData2Out,      // from registers to alu (or alu input mux)
    Address *exm_pcAluOut,                  // from PC to instruction memory
    Address *exm_branchAddrOut,          // from branch target adder to branch mux
    Instruction *exm_instrMemOut,        // output from instruction memory
    Address *exm_jCombineOut,            // from jump address combiner to jump mux
    RegDstSel *exm_regDstOut,               // is result register selected by rt or rd?
    Bool *exm_branchOut,                    // is this a branch instruction?
    JMuxSel *exm_jumpOut,                    // is this a branch instruction?
    Bool *exm_memReadOut,                   // read data memory?
    MemToRegSel *exm_memToRegOut,           // register output from alu or from data memory?
    Bool *exm_memWriteOut,                  // write data memory?
    Bool *exm_regWriteOut)
{
    *exm_zeroOut			= exm_zero;         
    *exm_aluResultOut		= exm_aluResult;    
    *exm_readData1Out		= exm_readData1;
    *exm_readData2Out		= exm_readData2;   
    *exm_pcAluOut		= exm_pcAlu;          
    *exm_branchAddrOut	= exm_branchAddr;  
    *exm_instrMemOut		= exm_instrMem;     
    *exm_jCombineOut		= exm_jCombine;     
    *exm_regDstOut		= exm_regDst;        
    *exm_branchOut		= exm_branch;        
    *exm_jumpOut		= exm_jump;        
    *exm_memReadOut		= exm_memRead;     
    *exm_memToRegOut	= exm_memToReg;   
    *exm_memWriteOut		= exm_memWrite;    
    *exm_regWriteOut		= exm_regWrite;
}
/* MEM/WB                                                                    */
/*...................................*/
void mem_wb(RegDstSel mwb_regDst,               // is result register selected by rt or rd?
    MemToRegSel mwb_memToReg,           // register output from alu or from data memory?
    Bool mwb_regWrite,                  // write result to registers?
    Data mwb_aluResult,                 // result of ALU operation
    Data mwb_readData1, Data mwb_readData2,      // from registers to alu (or alu input mux)
    Data mwb_readData,                  // from data mem to alu/mem mux
    Instruction mwb_instrMem,        // output from instruction memory
    Address mwb_jMux,                // from jump mux to PC
//and outs... :(
    RegDstSel *mwb_regDstOut,               // is result register selected by rt or rd?
    MemToRegSel *mwb_memToRegOut,           // register output from alu or from data memory?
    Bool *mwb_regWriteOut,                  // write result to registers?
    Data *mwb_aluResultOut,                 // result of ALU operation
    Data *mwb_readData1Out, Data *mwb_readData2Out,      // from registers to alu (or alu input mux)
    Data *mwb_readDataOut,                  // from data mem to alu/mem mux
    Instruction *mwb_instrMemOut,        // output from instruction memory
    Address *mwb_jMuxOut)
{
    *mwb_regDstOut		= mwb_regDst;   
    *mwb_memToRegOut	= mwb_memToReg; 
    *mwb_regWriteOut		= mwb_regWrite;    
    *mwb_aluResultOut		= mwb_aluResult;    
    *mwb_readData1Out	= mwb_readData1; 
    *mwb_readData2Out	= mwb_readData2;  
    *mwb_readDataOut		= mwb_readData;    
    *mwb_instrMemOut		= mwb_instrMem;    
    *mwb_jMuxOut		= mwb_jMux;
}
    
/*                                                                           */
/*******************************END PIPELINE REGS*********************/

/*****************************************************************************/
/*                                                                           */
/*  report                                                                   */
/*                                                                           */
/*      reports simulation state                                             */
/*                                                                           */
/*      inputs:                                                              */
/*          pc:  program counter                                             */
/*          registers:    register file                                      */
/*          data_memory:  data memory                                        */
/*                                                                           */
/*****************************************************************************/
void report(uint32_t pc, uint32_t registers[NUMREGS], uint32_t data_memory[DATAMEMSIZE])
{
	static uint32_t old_registers[NUMREGS];
	static uint32_t old_data_memory[DATAMEMSIZE];
	static iteration = 0;
	int i;

	if (iteration == 0) {
		printf("INITIALIZING REPORT FUNCTION\n");
		for (i = 0; i < NUMREGS; i++)
			old_registers[i] = registers[i];
		for (i = 0; i < DATAMEMSIZE; i++)
			old_data_memory[i] = data_memory[i];
	} else {
		
		printf("Modified registers:\n");
		for (i = 0; i < NUMREGS; i++)
			if (old_registers[i] != registers[i]) {
				printf("    R[%2d]: 0x%08x => 0x%08x\n", i, old_registers[i], registers[i]);
				old_registers[i] = registers[i];
			}

		printf("Modified memory locations:\n");
		for (i = 0; i < DATAMEMSIZE; i++)
			if (old_data_memory[i] != data_memory[i]) {
				printf("    Mem[0x%08x]: 0x%08x => 0x%08x\n", i, old_data_memory[i], data_memory[i]);
				old_data_memory[i] = data_memory[i];
			}
	}

	printf("\n\nCYCLE %2d\n", iteration);
	printf("PC:  0x%08x\n", pc);
	
	iteration++;
}
